/*
 *  $Id$
 * 
 *  Copyright 2007 Sun Microsystems, Inc. All rights reserved.
 *  SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package javafx.ui;
import javax.swing.*;
import javafx.ui.ScrollableWidget;
import javafx.ui.ListCell;
import javafx.ui.ListLayoutOrientation;
import javafx.ui.MultiSelection;
import javafx.ui.Border;
import javafx.ui.LineBorder;
import javafx.ui.Label;
import java.awt.MouseInfo;
import javax.swing.TransferHandler;

class ListDropEvent extends DropEvent {
    public attribute listIndex: Integer;
}
/**
 * A component that allows the user to select one or more objects from a
 * list. Encapsulates javax.swing.JList.
 */

public class ListBox extends ScrollableWidget {
    private attribute transferHandler: TransferHandler;
    private attribute selectedCell:ListCell?;
    private attribute updateGeneration: Number;
    private attribute selectionGeneration: Number;
    private operation acceptDrop(value):Boolean;
    private operation getDragValue(): *;
    private operation getDragText(): String;
    private operation setDropValue(value);
    private attribute dirty: Boolean;
    private attribute keyListener: <<java.awt.event.KeyListener>>;
    private attribute listMouseListener: <<java.awt.event.MouseListener>>;
    private attribute list: <<javax.swing.JList>>;
    private attribute selectionListener:<<javax.swing.event.ListSelectionListener>>;
    private attribute listeners:<<javax.swing.event.ListDataListener>>*;
    private attribute listmodel:<<javax.swing.ListModel>>;

    public attribute enableDND: Boolean;
    /** 
     * If true the list will not reflect changes to its cells attribute. 
     * When changed back to false synchronization will occur. Can be used 
     * as a performance optimizationwhen when doing large batch updates.
     */
    public attribute locked: Boolean?;

    /**
     * Defines the way list cells are layed out. Consider a <code>ListBox</code>
     * with four cells, this can be layed out in one of the following ways:
     * <pre>
     *   0
     *   1
     *   2
     *   3
     * </pre>
     * <pre>
     *   0  1
     *   2  3
     * </pre>
     * <pre>
     *   0  2
     *   1  3
     * </pre>
     * <p>
     * These correspond to the following values:
     *
     * <table border="1" 
     *  summary="Describes layouts VERTICAL, HORIZONTAL_WRAP, and VERTICAL_WRAP">
     *   <tr><th><p align="left">Value</p></th><th><p align="left">Description</p></th></tr>
     *   <tr><td><code>VERTICAL</code>
     *       <td>The cells should be layed out vertically in one column.
     *   <tr><td><code>HORIZONTAL_WRAP</code>
     *       <td>The cells should be layed out horizontally, wrapping to
     *           a new row as necessary.  The number
     *           of rows to use will either be defined by
     *           <code>visibleRowCount</code> if > 0, otherwise the
     *           number of rows will be determined by the width of the 
     *           <code>ListBox</code>.
     *   <tr><td><code>VERTICAL_WRAP</code>
     *       <td>The cells should be layed out vertically, wrapping to a
     *           new column as necessary.  The number
     *           of rows to use will either be defined by
     *           <code>visibleRowCount</code> if > 0, otherwise the
     *           number of rows will be determined by the height of the 
     *           <code>ListBox</code>.
     *  </table>
     * The default value of this property is <code>VERTICAL</code>.
     * <p>
     */
    public attribute layoutOrientation: ListLayoutOrientation?;
    /** Returns the index of the selected cell or -1 if no cell is selected. */
    public attribute selection: Number;
    /**
     * Returns the preferred number of visible rows.
     */
    public attribute visibleRowCount: Number?;
    /**
     * If present, sets the height of every cell in the list. 
     */
    public attribute fixedCellHeight: Number?;
    /**
     * If present, sets the width of every cell in the list. 
     */
    public attribute fixedCellWidth: Number?;
    /** The cells of this list. */
    public attribute cells: ListCell*;
    private attribute onSelectionChange: function(oldSelection:MultiSelection, newSelection:MultiSelection)?;
    /** 
     * <code>attribute action: function()?</code><br></br>
     * Optional handler called when the user double-clicks on a list cell or
     * presses the ENTER key.
     */
    public attribute action: function()?;
    /**
     * Optional handler called when the user drops an object into this list.
     */
    public attribute onDrop: function(e:ListDropEvent)?;
    /**
     * Optional filter for the types of objects that may be dropped into this
     * list.
     */
    public attribute dropType: Class?;
    /**
     * <code>attribute acceptDrop: function(value): Boolean</code><br></br>
     * Optional handler called when the user drops an object into this list.
     * If it returns false, the drop is rejected. 
     */
    public attribute canAcceptDrop: function(e:ListDropEvent):Boolean?;

    public operation locationToIndex(x:Number, y:Number);
    public operation indexToLocation(i:Number):XY;
}

operation ListBox.locationToIndex(x:Number, y:Number) {
    return list.locationToIndex(new <<java.awt.Point>>(x, y));
}

operation ListBox.indexToLocation(i) {
    var pt = list.indexToLocation(i);
    return XY {x: pt.x, y: pt.y};
}

attribute ListBox.selection = -1;

trigger on (ListBox.action = value) {
    if (value == null) {
        if (keyListener <> null) {
            list.removeKeyListener(keyListener);
        }
        if (listMouseListener <> null) {
            list.removeMouseListener(listMouseListener);
        }
    } else {
        if (keyListener == null) {
            var self = this;
            keyListener = new <<java.awt.event.KeyListener>>() {
                operation keyPressed(e:<<java.awt.event.KeyEvent>>) {
                    if (e.getKeyCode() == <<java.awt.event.KeyEvent>>.VK_ENTER) {
                        (self.action)();
                    }
                }
            };
            listMouseListener = new <<java.awt.event.MouseListener>>() {
                operation mouseClicked(e:<<java.awt.event.MouseEvent>>) {
                    if (e.getClickCount() == 2 and
                        <<javax.swing.SwingUtilities>>.isLeftMouseButton(e)) {
                        (self.action)();
                    }
                }
            };
        }
        list.addKeyListener(keyListener);
        list.addMouseListener(listMouseListener);
    }
}

trigger on (ListBox.visibleRowCount = n) {
    list.setVisibleRowCount(n);
    list.revalidate();
}

trigger on (ListBox.layoutOrientation = value) {
    list.setLayoutOrientation(value.id);
}


trigger on (ListBox.fixedCellHeight = n) {
    list.setFixedCellHeight(n);
}

trigger on (ListBox.fixedCellWidth = n) {
    list.setFixedCellWidth(n);
}

trigger on (ListBox.selection[old] = newValue) {
    if (not locked) {
        list.removeListSelectionListener(selectionListener);
        list.setSelectedIndex(newValue);
        if (newValue >= 0) {
            cells[old].selected = false;
            cells[newValue].selected = true;
            list.ensureIndexIsVisible(newValue);
        }
        list.addListSelectionListener(selectionListener);
    }
}

operation ListBox.getDragValue() {
    var result = [];
    for (i in selection) {
        insert cells[i].value into result;
    }
    return result;
}

operation ListBox.getDragText() {
    var selectedCells:ListCell* = [];
    for (i in selection) {
        insert cells[i] into selectedCells;
    }
    // fix me...
    return if selectedCells[0].dragText <> null then
        selectedCells[0].dragText else selectedCells[0].text;
}

operation ListBox.setDropValue(value) {
    if (this.onDrop <> null) {
        var info = MouseInfo.getPointerInfo();
        var location = list.getLocationOnScreen();
        var p = info.getLocation();
        p.x -= location.x;
        p.y -= location.y;
        var i = locationToIndex(p.x, p.y);
        var e = ListDropEvent {
            x: p.x
            y: p.y
            transferData: value
            listIndex: i
        };
        (this.onDrop)(e);
    }
}

operation ListBox.acceptDrop(value) {
    if (this.canAcceptDrop <> null) {
        var info = MouseInfo.getPointerInfo();
        var location = list.getLocationOnScreen();
        var p = info.getLocation();
        p.x -= location.x;
        p.y -= location.y;
        var i = locationToIndex(p.x, p.y);
        var e = ListDropEvent {
            x: p.x
            y: p.y
            transferData: value
            listIndex: i
        };
        return (this.canAcceptDrop)(e);
    }
    return onDrop <> null;
}

trigger on ListBox.enableDND = value {
    if (value) {
        if (transferHandler == null) {
            var self = this;
            context:<<net.java.javafx.ui.UIContext>>.addTransferHandler(list,
                                                                    dropType,
                                                                    new <<net.java.javafx.ui.ValueGetter>>() {
                                                                    operation get() {
                                                                        return self.getDragValue();
                                                                    }
                                                                },
                                                                new <<net.java.javafx.ui.ValueSetter>>() {
                                                                    operation set(value) {
                                                                        self.setDropValue(value);
                                                                        
                                                                    }
                                                                },
                                                                new <<net.java.javafx.ui.ValueAcceptor>>() {
                                                                    operation accept(value) {
                                                                        return self.acceptDrop(value);
                                                                    }
                                                                },
                                                                new <<net.java.javafx.ui.VisualRepresentation>>() {
                                                                    operation getComponent(value) {
                                                                        var label = Label {
                                                                            opaque: true
                                                                            border: LineBorder {
                                                                                lineColor: black
                                                                            }
                                                                            text: self.getDragText()
                                                                };
                                                                        return label.getComponent();
                                                                    }
                                                                });
            transferHandler = list.getTransferHandler();
        }
        list.setDragEnabled(true);
    } else {
        list.setDragEnabled(false);
    }
}

operation ListBox.createView():<<javax.swing.JComponent>> {

    //do later {
        locked = false;
    //}
    return list;
}

trigger on (new ListBox) {
    layoutOrientation = VERTICAL:ListLayoutOrientation;
    list = new <<javax.swing.JList>>;
    // fix me...
    list.setSelectionMode(<<javax.swing.ListSelectionModel>>.SINGLE_SELECTION);
    selection = -1;
    var self = this;
    listmodel = new <<javax.swing.ListModel>>() {
        operation getSize() {
            return sizeof self.cells;
        }
        operation getElementAt(i:Number) {
            return self.cells[i];
        }
        operation addListDataListener(l:<<javax.swing.event.ListDataListener>>) {
            insert l into self.listeners;
        }
        operation removeListDataListener(l:<<javax.swing.event.ListDataListener>>) {
            delete self.listeners[. == l];
        }
    };
    list.setCellRenderer(new <<javax.swing.ListCellRenderer>>() {
                             operation getListCellRendererComponent(jlist:<<javax.swing.JList>>, obj,
                                                                    i:Number, selected:Boolean, focused:Boolean) {
                                 var value = (ListCell)obj;
                                 var label = (<<javax.swing.JLabel>>) context:<<net.java.javafx.ui.UIContext>>.getListCellRendererComponent(jlist,
                                                                                                      value.text,
                                                                                                      i, selected, focused,
                                                                                                      value.toolTipText);
                                if (value.border <> null) {
                                     label.setBorder(value.border.getBorder());
                                }
                                if (value.horizontalAlignment <> null) {
                                    label.setHorizontalAlignment(value.horizontalAlignment.id);
                                } 
                                if (value.verticalAlignment <> null) {
                                    label.setVerticalAlignment(value.verticalAlignment.id);
                                }
                                //label.setOpaque(self.list.isOpaque());
                                return label;
                             }
                             operation equals(o) {
                                 return o == this;
                             }
                         });
    <<javax.swing.ToolTipManager>>.sharedInstance().registerComponent(list);
    list.setModel(listmodel);
    selectionListener = new <<javax.swing.event.ListSelectionListener>>() {
                                  operation valueChanged(e:<<javax.swing.event.ListSelectionEvent>>) {
                                      if (not e.getValueIsAdjusting()) {
                                              self.selection = self.list.getSelectedIndex();

                                      }
                                  }
                              };
    locked = true;
}


trigger on (ListBox.locked[old] = value) {
    if (not value) {
        list.removeListSelectionListener(selectionListener);
        if (dirty) {
            dirty = false;
            delete listeners;
            list.setModel(new <<javax.swing.DefaultListModel>>()) ;
            list.setModel(listmodel);
            list.setSelectedIndex(selection);
            list.ensureIndexIsVisible(selection);
        }
        list.addListSelectionListener(selectionListener);
    } else {
        list.removeListSelectionListener(selectionListener);
    }
}

trigger on (insert cell into ListBox.cells) {
    cell.listbox = this;
    cell.cacheGeneration = ++updateGeneration;
    cell.myIndex = indexof cell;
    if (not locked) {
        var e:<<javax.swing.event.ListDataEvent>>;
        var i = indexof cell;
        var selected = i in selection or sizeof cells == 1;
        e = new <<javax.swing.event.ListDataEvent>>(list, e.INTERVAL_ADDED, i, i);
        var ls:<<javax.swing.event.ListDataListener>>* = [];
        insert listeners into ls;
        for (j in ls) {
            j.intervalAdded(e);
        }
        if (selected or cell.selected) {
            list.removeListSelectionListener(selectionListener);
            list.setSelectedIndex(i);
            selection = i;
            list.addListSelectionListener(selectionListener);
        }
    } else {
        dirty = true;
    }
}

trigger on (delete cell from ListBox.cells) {
    ++updateGeneration;
    if (not locked) {
        var e:<<javax.swing.event.ListDataEvent>>;
        var i = indexof cell;
        var s = selection;
        e = new <<javax.swing.event.ListDataEvent>>(list, e.INTERVAL_REMOVED, i, i);
        var ls:<<javax.swing.event.ListDataListener>>* = [];
        insert listeners into ls;
        for (j in ls) {
            j.intervalRemoved(e);
        }
        if (i == s) {
            if (s >= sizeof cells) {
                s = sizeof cells -1;
            }
            if (s >= 0) {
               list.setSelectedIndex(s);
            }
        }
    } else {
        dirty = true;
    }
}
